# Copyright (c) 2021 J. Lewis Muir
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

require 'csv'
require 'fileutils'
require 'time'
require 'tmpdir'
require 'stringio'

require 'test/unit'

require_relative 'imca_shipment_tool_app'

#
# Tests IMCAShipmentToolApp.
#
class IMCAShipmentToolAppTest < Test::Unit::TestCase
  @@unix_like_chmod_mutex = Mutex.new
  @@unix_like_chmod = nil

  def self.create_app_context
    {
      prog_name: nil,
      argv: [],
      stdin: StringIO.new('', 'r'),
      stdout: StringIO.new('', 'w'),
      stderr: StringIO.new('', 'w'),
      env: {}
    }
  end

  def self.unix_like_chmod?
    @@unix_like_chmod_mutex.synchronize do
      return @@unix_like_chmod unless @@unix_like_chmod.nil?
      @@unix_like_chmod = Dir.mktmpdir(File.basename($0)) do |dir|
        f_pathname = File.join(dir, 'foo')
        File.open(f_pathname, 'w') { |f| }
        File.chmod(0644, f_pathname)
        File.chmod(0640, f_pathname)
        File.stat(f_pathname).mode & 06777 == 0640
      end
    end
  end

  def test_help_option
    ['-h', '--help'].each do |opt|
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = [opt]
      e = assert_raise(SystemExit) { IMCAShipmentToolApp.new(**app_ctx).run }
      assert_equal(1, e.status)
      assert_match(/^usage: /, app_ctx[:stderr].string)
    end
  end

  def test_no_command
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = []
    e = assert_raise(SystemExit) { IMCAShipmentToolApp.new(**app_ctx).run }
    assert_equal(1, e.status)
    assert_match(/^usage: /, app_ctx[:stderr].string)
  end

  def test_invalid_command
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['foo']
    e = assert_raise(SystemExit) { IMCAShipmentToolApp.new(**app_ctx).run }
    assert_equal(1, e.status)
    assert_match(/^usage: /, app_ctx[:stderr].string)
  end

  def test_sort_help_option
    ['-h', '--help'].each do |opt|
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', opt]
      e = assert_raise(SystemExit) { IMCAShipmentToolApp.new(**app_ctx).run }
      assert_equal(1, e.status)
      assert_match(/^usage: /, app_ctx[:stderr].string)
    end
  end

  def test_sort_invalid_option
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort', '--foo']
    e = assert_raise(SystemExit) { IMCAShipmentToolApp.new(**app_ctx).run }
    assert_equal(1, e.status)
    assert_match(/^usage: /, app_ctx[:stderr].string)
  end

  def test_sort_verbose_option
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      FileUtils.touch(f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-v', f1_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_match(
        Regexp.new(Regexp.quote(f1_pathname)),
        app_ctx[:stdout].string
      )
    end
  end

  def test_sort_nonexistent_reference_file
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      e = assert_raise(SystemExit) { IMCAShipmentToolApp.new(**app_ctx).run }
      assert_equal(1, e.status)
      assert_not_empty(app_ctx[:stderr].string)
    end
  end

  def test_sort_nonreadable_reference_file
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      FileUtils.touch(r1_pathname)
      File.chmod(0, r1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      e = assert_raise(SystemExit) { IMCAShipmentToolApp.new(**app_ctx).run }
      assert_equal(1, e.status)
      assert_not_empty(app_ctx[:stderr].string)
    end
  end

  def test_sort_stdio_empty_reference_file
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      FileUtils.touch(r1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      app_ctx[:stdin] = StringIO.new("a,1\n", 'r')
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal([['a', '1']], csv)
    end
  end

  def test_sort_stdio_zero_length
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    assert_empty(app_ctx[:stdout].string)
  end

  def test_sort_file_zero_length
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      FileUtils.touch(f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', f1_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_equal(0, File.size(f1_pathname))
    end
  end

  def test_sort_stdio_invalid_csv_too_few_columns
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new("foo\n", 'r')
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(1, status)
    assert_match(/Too few columns/, app_ctx[:stderr].string)
  end

  def test_sort_file_invalid_csv_too_few_columns
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(f1_pathname, "foo\n", mode: 'w')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', f1_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(/Too few columns/, app_ctx[:stderr].string)
    end
  end

  def test_sort_one_pin
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "c,1\n" \
      "a,1\n" \
      "b,1\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['a', '1'],
        ['b', '1'],
        ['c', '1']
      ],
      csv
    )
  end

  # Duplicate pins are invalid, but test for a stable sort anyway.
  def test_sort_one_pin_dups_stable
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "c,1,x06\n" \
      "c,1,x05\n" \
      "a,1,x04\n" \
      "a,1,x03\n" \
      "b,1,x02\n" \
      "b,1,x01\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['a', '1', 'x04'],
        ['a', '1', 'x03'],
        ['b', '1', 'x02'],
        ['b', '1', 'x01'],
        ['c', '1', 'x06'],
        ['c', '1', 'x05']
      ],
      csv
    )
  end

  def test_sort_multiple_pins
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "c,3\n" \
      "c,1\n" \
      "c,2\n" \
      "a,3\n" \
      "a,1\n" \
      "a,2\n" \
      "b,3\n" \
      "b,1\n" \
      "b,2\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['a', '1'],
        ['a', '2'],
        ['a', '3'],
        ['b', '1'],
        ['b', '2'],
        ['b', '3'],
        ['c', '1'],
        ['c', '2'],
        ['c', '3']
      ],
      csv
    )
  end

  # Duplicate pins are invalid, but test for a stable sort anyway.
  def test_sort_multiple_pins_dups_stable
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "c,3,x12\n" \
      "c,1,x11\n" \
      "c,2,x10\n" \
      "c,2,x09\n" \
      "a,3,x08\n" \
      "a,1,x07\n" \
      "a,1,x06\n" \
      "a,2,x05\n" \
      "b,3,x04\n" \
      "b,3,x03\n" \
      "b,1,x02\n" \
      "b,2,x01\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['a', '1', 'x07'],
        ['a', '1', 'x06'],
        ['a', '2', 'x05'],
        ['a', '3', 'x08'],
        ['b', '1', 'x02'],
        ['b', '2', 'x01'],
        ['b', '3', 'x04'],
        ['b', '3', 'x03'],
        ['c', '1', 'x11'],
        ['c', '2', 'x10'],
        ['c', '2', 'x09'],
        ['c', '3', 'x12']
      ],
      csv
    )
  end

  def test_sort_mixed_containers_pins
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "c,3\n" \
      "a,1\n" \
      "b,2\n" \
      "c,2\n" \
      "b,1\n" \
      "a,3\n" \
      "b,3\n" \
      "c,1\n" \
      "a,2\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['a', '1'],
        ['a', '2'],
        ['a', '3'],
        ['b', '1'],
        ['b', '2'],
        ['b', '3'],
        ['c', '1'],
        ['c', '2'],
        ['c', '3']
      ],
      csv
    )
  end

  # Duplicate pins are invalid, but test for a stable sort anyway.
  def test_sort_mixed_containers_pins_dups_stable
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "c,3,x12\n" \
      "c,3,x11\n" \
      "a,1,x10\n" \
      "b,2,x09\n" \
      "c,2,x08\n" \
      "b,1,x07\n" \
      "b,1,x06\n" \
      "a,3,x05\n" \
      "b,3,x04\n" \
      "c,1,x03\n" \
      "a,2,x02\n" \
      "a,2,x01\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['a', '1', 'x10'],
        ['a', '2', 'x02'],
        ['a', '2', 'x01'],
        ['a', '3', 'x05'],
        ['b', '1', 'x07'],
        ['b', '1', 'x06'],
        ['b', '2', 'x09'],
        ['b', '3', 'x04'],
        ['c', '1', 'x03'],
        ['c', '2', 'x08'],
        ['c', '3', 'x12'],
        ['c', '3', 'x11']
      ],
      csv
    )
  end

  # An empty reference file is invalid, but test for a stable sort anyway.
  def test_sort_one_pin_with_empty_reference_file_stable
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      FileUtils.touch(r1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      app_ctx[:stdin] = StringIO.new(
        "c,1\n" \
        "a,1\n" \
        "b,1\n",
        'r'
      )
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          ['c', '1'],
          ['a', '1'],
          ['b', '1']
        ],
        csv
      )
    end
  end

  # An empty reference file is invalid, but test for a stable sort anyway.
  def test_sort_multiple_pins_with_empty_reference_file_stable
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      FileUtils.touch(r1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      app_ctx[:stdin] = StringIO.new(
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n",
        'r'
      )
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          ['c', '1'],
          ['c', '2'],
          ['c', '3'],
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3']
        ],
        csv
      )
    end
  end

  # An empty reference file is invalid, but test for a stable sort anyway.
  def test_sort_mixed_containers_pins_with_empty_reference_file_stable
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      FileUtils.touch(r1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      app_ctx[:stdin] = StringIO.new(
        "c,3\n" \
        "a,1\n" \
        "b,2\n" \
        "c,2\n" \
        "b,1\n" \
        "a,3\n" \
        "b,3\n" \
        "c,1\n" \
        "a,2\n",
        'r'
      )
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          ['c', '1'],
          ['c', '2'],
          ['c', '3'],
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3']
        ],
        csv
      )
    end
  end

  def test_sort_header_detect
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "Puck,Pin\n" \
      "C,3\n" \
      "C,1\n" \
      "C,2\n" \
      "A,3\n" \
      "A,1\n" \
      "A,2\n" \
      "B,3\n" \
      "B,1\n" \
      "B,2\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['Puck', 'Pin'],
        ['A', '1'],
        ['A', '2'],
        ['A', '3'],
        ['B', '1'],
        ['B', '2'],
        ['B', '3'],
        ['C', '1'],
        ['C', '2'],
        ['C', '3']
      ],
      csv
    )
  end

  def test_sort_header_detect_no_data
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "Puck,Pin\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['Puck', 'Pin'],
      ],
      csv
    )
  end

  def test_sort_multiple_pins_with_reference_file
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      IO.write(
        r1_pathname,
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      app_ctx[:stdin] = StringIO.new(
        "a,1\n" \
        "a,2\n" \
        "a,3\n" \
        "b,1\n" \
        "b,2\n" \
        "b,3\n" \
        "c,1\n" \
        "c,2\n" \
        "c,3\n",
        'r'
      )
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          ['c', '1'],
          ['c', '2'],
          ['c', '3'],
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3']
        ],
        csv
      )
    end
  end

  def test_sort_multiple_files
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(
        f1_pathname,
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n",
        mode: 'w'
      )
      f2_pathname = File.join(dir, 'f2.csv')
      IO.write(
        f2_pathname,
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', f1_pathname, f2_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          # f1
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          ['c', '1'],
          ['c', '2'],
          ['c', '3'],
          # f2
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          ['c', '1'],
          ['c', '2'],
          ['c', '3']
        ],
        csv
      )
    end
  end

  def test_sort_multiple_files_with_reference_file
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      IO.write(
        r1_pathname,
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n",
        mode: 'w'
      )
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(
        f1_pathname,
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n",
        mode: 'w'
      )
      f2_pathname = File.join(dir, 'f2.csv')
      IO.write(
        f2_pathname,
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n" \
        "c,3\n" \
        "c,1\n" \
        "c,2\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname, f1_pathname, f2_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          # f1
          ['c', '1'],
          ['c', '2'],
          ['c', '3'],
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          # f2
          ['c', '1'],
          ['c', '2'],
          ['c', '3'],
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3']
        ],
        csv
      )
    end
  end

  def test_sort_multiple_files_in_place
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(
        f1_pathname,
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n",
        mode: 'w'
      )
      f2_pathname = File.join(dir, 'f2.csv')
      IO.write(
        f2_pathname,
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-i', f1_pathname, f2_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      assert_empty(app_ctx[:stdout].string)
      f1_csv = CSV.read(f1_pathname)
      assert_equal(
        [
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          ['c', '1'],
          ['c', '2'],
          ['c', '3']
        ],
        f1_csv
      )
      f2_csv = CSV.read(f2_pathname)
      assert_equal(
        [
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          ['c', '1'],
          ['c', '2'],
          ['c', '3']
        ],
        f2_csv
      )
    end
  end

  def test_sort_multiple_files_skips_nonexistent_files
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      f2_pathname = File.join(dir, 'f2.csv')
      IO.write(
        f2_pathname,
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', f1_pathname, f2_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(
        Regexp.new(": #{Regexp.quote(f1_pathname)}: "),
        app_ctx[:stderr].string
      )
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          ['c', '1'],
          ['c', '2'],
          ['c', '3']
        ],
        csv
      )
    end
  end

  def test_sort_multiple_files_in_place_skips_nonexistent_files
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      f2_pathname = File.join(dir, 'f2.csv')
      IO.write(
        f2_pathname,
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-i', f1_pathname, f2_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(
        Regexp.new(": #{Regexp.quote(f1_pathname)}: "),
        app_ctx[:stderr].string
      )
      assert_empty(app_ctx[:stdout].string)
      assert_path_not_exist(f1_pathname)
      f2_csv = CSV.read(f2_pathname)
      assert_equal(
        [
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          ['c', '1'],
          ['c', '2'],
          ['c', '3']
        ],
        f2_csv
      )
    end
  end

  def test_sort_multiple_files_in_place_skips_if_tmp_file_exists
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(
        f1_pathname,
        "a,3\n" \
        "a,1\n" \
        "a,2\n" \
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n",
        mode: 'w'
      )
      f1_tmp_pathname = File.join(dir, 'f1.csv.tmp')
      FileUtils.touch(f1_tmp_pathname)
      f2_pathname = File.join(dir, 'f2.csv')
      IO.write(
        f2_pathname,
        "c,3\n" \
        "c,1\n" \
        "c,2\n" \
        "b,3\n" \
        "b,1\n" \
        "b,2\n" \
        "a,3\n" \
        "a,1\n" \
        "a,2\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-i', f1_pathname, f2_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(
        Regexp.new(": #{Regexp.quote(f1_tmp_pathname)}: "),
        app_ctx[:stderr].string
      )
      assert_empty(app_ctx[:stdout].string)
      f1_csv = CSV.read(f1_pathname)
      assert_equal(
        [
          ['a', '3'],
          ['a', '1'],
          ['a', '2'],
          ['c', '3'],
          ['c', '1'],
          ['c', '2'],
          ['b', '3'],
          ['b', '1'],
          ['b', '2']
        ],
        f1_csv
      )
      f2_csv = CSV.read(f2_pathname)
      assert_equal(
        [
          ['a', '1'],
          ['a', '2'],
          ['a', '3'],
          ['b', '1'],
          ['b', '2'],
          ['b', '3'],
          ['c', '1'],
          ['c', '2'],
          ['c', '3']
        ],
        f2_csv
      )
    end
  end

  def test_sort_file_in_place_preserve_attributes
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(f1_pathname, "a,1\n", mode: 'w')
      orig_fs1 = File.stat(f1_pathname)
      mode = 0604
      File.chmod(mode, f1_pathname)
      gid = nil
      begin
        Process.groups.find { |x| x != orig_fs1.gid }
      rescue NotImplementedError
        # Process.groups is not implemented on some platforms.
      end
      File.chown(nil, gid, f1_pathname) if gid
      atime = Time.parse('1999-01-02T03:04:05.678')
      mtime = Time.parse('1998-02-03T04:05:06.789')
      File.utime(atime, mtime, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-i', f1_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      if self.class.unix_like_chmod?
        assert_equal(mode, fs1.mode & 06777)
      end
      assert_equal(gid, fs1.gid) if gid
      assert_not_equal(atime, fs1.atime)
      assert_not_equal(mtime, fs1.mtime)
    end
  end

  def test_sort_file_in_place_preserve_attributes_set_uid_on_exec
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(f1_pathname, "a,1\n", mode: 'w')
      mode = 04604
      File.chmod(mode, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-i', f1_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      assert_equal(mode, fs1.mode & 06777)
    end
  end

  def test_sort_file_in_place_preserve_attributes_set_gid_on_exec
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(f1_pathname, "a,1\n", mode: 'w')
      mode = 02604
      File.chmod(mode, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-i', f1_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      assert_equal(mode, fs1.mode & 06777)
    end
  end

  def test_sort_file_in_place_preserve_attributes_set_uidgid_on_exec
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f1.csv')
      IO.write(f1_pathname, "a,1\n", mode: 'w')
      mode = 06604
      File.chmod(mode, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-i', f1_pathname]
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      assert_equal(mode, fs1.mode & 06777)
    end
  end

  def test_sort_real_world_data
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['sort']
    app_ctx[:stdin] = StringIO.new(
      "Puck,Pin,Project,Priority,Mode,Notes,Strategy,Contact,Expected Space Group,Expected Cell Dimensions,Expected Resolution,Minimum Collection Resolution,Recipe,Exposure Time,Image Width,Total Phi,Attenuation,Aperture,Detector Distance,Frame Names,Observed Resolution,Staff Comments,Status,done,returned\n" \
      "Blue54,1,e,,Auto,,,,P212121,67 82 276,2.1,2.5,,,0.25,180,,10,350,image_,,,queued,,\n" \
      "Blue54,2,e,4,,,,Luke,P2(1)2(1)2(1),67 82 276,,1.7,luke-180.rcp,,,,,,,,1.7,Processing failed,collected,,\n" \
      "Blue54,3,f,,Auto,,,,P222,,2.1,,,0.04,,180,95,,350,image_,2.1,,collected,,\n" \
      "Blue54,4,g,4,Auto,,,Luke,,,2.1,2.5,,0.04,0.25,180,75,,350,image_,,,queued,,\n" \
      "Blue54,5,h,99,Auto,,,Luke,P222,,2.2,2.5,,0.04,0.25,180,95,,400,image_,,,queued,,\n" \
      "Blue53,1,a,1,Manual,Tricky,Do best you can,Luke,C2,\"143.734, 67.095, 76.899, 90, 110.45, 90\",3.5,4,luke-360.rcp,,,,,,,,1.9,,collected,2019-06-09 08:23:01,\n" \
      "Blue53,2,a,1,Manual,Very tricky,New crystals,Luke,C2,140 65 75 90 110 90,2.4,3.5,,0.1,0.25,,95,5,250,image_,2.8,Badly split; no good spot,collected,2019-06-09 08:26:01,\n" \
      "Blue53,3,b,3,Auto,Routine,SeMet,Luke,P2,52.4 39.8 65.0 108.5,1.5,1.7,,0.04,0.25,360,,10,300,,,Diff. center cutoff not met,skipped,2019-06-09 08:29:01,\n" \
      "Blue53,4,c,3,Auto,Rods,Native,Luke,P21,39 69.2 60 90 105.3,1.5,1.7,,0.04,0.25,360,95,20,,image_,1.6,,collected,,\n" \
      "Blue53,5,d,8,,Plates,,Luke,C222,280 45 112 102 90,1.5,1.7,,0.04,0.25,360,95,50,300,image_,,,received,,\n",
      'r'
    )
    status = IMCAShipmentToolApp.new(**app_ctx).run
    assert_equal(0, status)
    assert_empty(app_ctx[:stderr].string)
    csv = CSV.parse(app_ctx[:stdout].string)
    assert_equal(
      [
        ['Puck', 'Pin', 'Project', 'Priority', 'Mode', 'Notes', 'Strategy', 'Contact', 'Expected Space Group', 'Expected Cell Dimensions', 'Expected Resolution', 'Minimum Collection Resolution', 'Recipe', 'Exposure Time', 'Image Width', 'Total Phi', 'Attenuation', 'Aperture', 'Detector Distance', 'Frame Names', 'Observed Resolution', 'Staff Comments', 'Status', 'done', 'returned'],
        ['Blue53', '1', 'a', '1', 'Manual', 'Tricky', 'Do best you can', 'Luke', 'C2', '143.734, 67.095, 76.899, 90, 110.45, 90', '3.5', '4', 'luke-360.rcp', nil, nil, nil, nil, nil, nil, nil, '1.9', nil, 'collected', '2019-06-09 08:23:01', nil],
        ['Blue53', '2', 'a', '1', 'Manual', 'Very tricky', 'New crystals', 'Luke', 'C2', '140 65 75 90 110 90', '2.4', '3.5', nil, '0.1', '0.25', nil, '95', '5', '250', 'image_', '2.8', 'Badly split; no good spot', 'collected', '2019-06-09 08:26:01', nil],
        ['Blue53', '3', 'b', '3', 'Auto', 'Routine', 'SeMet', 'Luke', 'P2', '52.4 39.8 65.0 108.5', '1.5', '1.7', nil, '0.04', '0.25', '360', nil, '10', '300', nil, nil, 'Diff. center cutoff not met', 'skipped', '2019-06-09 08:29:01', nil],
        ['Blue53', '4', 'c', '3', 'Auto', 'Rods', 'Native', 'Luke', 'P21', '39 69.2 60 90 105.3', '1.5', '1.7', nil, '0.04', '0.25', '360', '95', '20', nil, 'image_', '1.6', nil, 'collected', nil, nil],
        ['Blue53', '5', 'd', '8', nil, 'Plates', nil, 'Luke', 'C222', '280 45 112 102 90', '1.5', '1.7', nil, '0.04', '0.25', '360', '95', '50', '300', 'image_', nil, nil, 'received', nil, nil],
        ['Blue54', '1', 'e', nil, 'Auto', nil, nil, nil, 'P212121', '67 82 276', '2.1', '2.5', nil, nil, '0.25', '180', nil, '10', '350', 'image_', nil, nil, 'queued', nil, nil],
        ['Blue54', '2', 'e', '4', nil, nil, nil, 'Luke', 'P2(1)2(1)2(1)', '67 82 276', nil, '1.7', 'luke-180.rcp', nil, nil, nil, nil, nil, nil, nil, '1.7', 'Processing failed', 'collected', nil, nil],
        ['Blue54', '3', 'f', nil, 'Auto', nil, nil, nil, 'P222', nil, '2.1', nil, nil, '0.04', nil, '180', '95', nil, '350', 'image_', '2.1', nil, 'collected', nil, nil],
        ['Blue54', '4', 'g', '4', 'Auto', nil, nil, 'Luke', nil, nil, '2.1', '2.5', nil, '0.04', '0.25', '180', '75', nil, '350', 'image_', nil, nil, 'queued', nil, nil],
        ['Blue54', '5', 'h', '99', 'Auto', nil, nil, 'Luke', 'P222', nil, '2.2', '2.5', nil, '0.04', '0.25', '180', '95', nil, '400', 'image_', nil, nil, 'queued', nil, nil]
      ],
      csv
    )
  end

  def test_sort_real_world_data_with_reference_file
    Dir.mktmpdir(File.basename($0)) do |dir|
      r1_pathname = File.join(dir, 'r1.csv')
      IO.write(
        r1_pathname,
        "Puck,Pin,Project,Priority,Mode,Notes,Strategy,Contact,Expected Space Group,Expected Cell Dimensions,Expected Resolution,Minimum Collection Resolution,Recipe,Exposure Time,Image Width,Total Phi,Attenuation,Aperture,Detector Distance,Frame Names,Observed Resolution,Staff Comments,Status,done,returned\n" \
        "Blue54,1,e,,Auto,,,,P212121,67 82 276,2.1,2.5,,,0.25,180,,10,350,image_,,,,,\n" \
        "Blue54,2,e,4,,,,Luke,P2(1)2(1)2(1),67 82 276,,1.7,luke-180.rcp,,,,,,,,,,,,\n" \
        "Blue54,3,f,,Auto,,,,P222,,2.1,,,0.04,,180,95,,350,image_,,,,,\n" \
        "Blue54,4,g,4,Auto,,,Luke,,,2.1,2.5,,0.04,0.25,180,75,,350,image_,,,,,\n" \
        "Blue54,5,h,99,Auto,,,Luke,P222,,2.2,2.5,,0.04,0.25,180,95,,400,image_,,,,,\n" \
        "Blue53,1,a,1,Manual,Tricky,Do best you can,Luke,C2,\"143.734, 67.095, 76.899, 90, 110.45, 90\",3.5,4,luke-360.rcp,,,,,,,,,,,,\n" \
        "Blue53,2,a,1,Manual,Very tricky,New crystals,Luke,C2,140 65 75 90 110 90,2.4,3.5,,0.1,0.25,,95,5,250,image_,,,,,\n" \
        "Blue53,3,b,3,Auto,Routine,SeMet,Luke,P2,52.4 39.8 65.0 108.5,1.5,1.7,,0.04,0.25,360,,10,300,,,,,,\n" \
        "Blue53,4,c,3,Auto,Rods,Native,Luke,P21,39 69.2 60 90 105.3,1.5,1.7,,0.04,0.25,360,95,20,,image_,,,,,\n" \
        "Blue53,5,d,8,,Plates,,Luke,C222,280 45 112 102 90,1.5,1.7,,0.04,0.25,360,95,50,300,image_,,,,,\n",
        mode: 'w'
      )
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['sort', '-r', r1_pathname]
      app_ctx[:stdin] = StringIO.new(
        "Puck,Pin,Project,Priority,Mode,Notes,Strategy,Contact,Expected Space Group,Expected Cell Dimensions,Expected Resolution,Minimum Collection Resolution,Recipe,Exposure Time,Image Width,Total Phi,Attenuation,Aperture,Detector Distance,Frame Names,Observed Resolution,Staff Comments,Status,done,returned\n" \
        "Blue53,1,a,1,Manual,Tricky,Do best you can,Luke,C2,\"143.734, 67.095, 76.899, 90, 110.45, 90\",3.5,4,luke-360.rcp,,,,,,,,1.9,,collected,2019-06-09 08:23:01,\n" \
        "Blue53,2,a,1,Manual,Very tricky,New crystals,Luke,C2,140 65 75 90 110 90,2.4,3.5,,0.1,0.25,,95,5,250,image_,2.8,Badly split; no good spot,collected,2019-06-09 08:26:01,\n" \
        "Blue53,3,b,3,Auto,Routine,SeMet,Luke,P2,52.4 39.8 65.0 108.5,1.5,1.7,,0.04,0.25,360,,10,300,,,Diff. center cutoff not met,skipped,2019-06-09 08:29:01,\n" \
        "Blue53,4,c,3,Auto,Rods,Native,Luke,P21,39 69.2 60 90 105.3,1.5,1.7,,0.04,0.25,360,95,20,,image_,1.6,,collected,,\n" \
        "Blue53,5,d,8,,Plates,,Luke,C222,280 45 112 102 90,1.5,1.7,,0.04,0.25,360,95,50,300,image_,,,received,,\n" \
        "Blue54,1,e,,Auto,,,,P212121,67 82 276,2.1,2.5,,,0.25,180,,10,350,image_,,,queued,,\n" \
        "Blue54,2,e,4,,,,Luke,P2(1)2(1)2(1),67 82 276,,1.7,luke-180.rcp,,,,,,,,1.7,Processing failed,collected,,\n" \
        "Blue54,3,f,,Auto,,,,P222,,2.1,,,0.04,,180,95,,350,image_,2.1,,collected,,\n" \
        "Blue54,4,g,4,Auto,,,Luke,,,2.1,2.5,,0.04,0.25,180,75,,350,image_,,,queued,,\n" \
        "Blue54,5,h,99,Auto,,,Luke,P222,,2.2,2.5,,0.04,0.25,180,95,,400,image_,,,queued,,\n",
        'r'
      )
      status = IMCAShipmentToolApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_empty(app_ctx[:stderr].string)
      csv = CSV.parse(app_ctx[:stdout].string)
      assert_equal(
        [
          ['Puck', 'Pin', 'Project', 'Priority', 'Mode', 'Notes', 'Strategy', 'Contact', 'Expected Space Group', 'Expected Cell Dimensions', 'Expected Resolution', 'Minimum Collection Resolution', 'Recipe', 'Exposure Time', 'Image Width', 'Total Phi', 'Attenuation', 'Aperture', 'Detector Distance', 'Frame Names', 'Observed Resolution', 'Staff Comments', 'Status', 'done', 'returned'],
          ['Blue54', '1', 'e', nil, 'Auto', nil, nil, nil, 'P212121', '67 82 276', '2.1', '2.5', nil, nil, '0.25', '180', nil, '10', '350', 'image_', nil, nil, 'queued', nil, nil],
          ['Blue54', '2', 'e', '4', nil, nil, nil, 'Luke', 'P2(1)2(1)2(1)', '67 82 276', nil, '1.7', 'luke-180.rcp', nil, nil, nil, nil, nil, nil, nil, '1.7', 'Processing failed', 'collected', nil, nil],
          ['Blue54', '3', 'f', nil, 'Auto', nil, nil, nil, 'P222', nil, '2.1', nil, nil, '0.04', nil, '180', '95', nil, '350', 'image_', '2.1', nil, 'collected', nil, nil],
          ['Blue54', '4', 'g', '4', 'Auto', nil, nil, 'Luke', nil, nil, '2.1', '2.5', nil, '0.04', '0.25', '180', '75', nil, '350', 'image_', nil, nil, 'queued', nil, nil],
          ['Blue54', '5', 'h', '99', 'Auto', nil, nil, 'Luke', 'P222', nil, '2.2', '2.5', nil, '0.04', '0.25', '180', '95', nil, '400', 'image_', nil, nil, 'queued', nil, nil],
          ['Blue53', '1', 'a', '1', 'Manual', 'Tricky', 'Do best you can', 'Luke', 'C2', '143.734, 67.095, 76.899, 90, 110.45, 90', '3.5', '4', 'luke-360.rcp', nil, nil, nil, nil, nil, nil, nil, '1.9', nil, 'collected', '2019-06-09 08:23:01', nil],
          ['Blue53', '2', 'a', '1', 'Manual', 'Very tricky', 'New crystals', 'Luke', 'C2', '140 65 75 90 110 90', '2.4', '3.5', nil, '0.1', '0.25', nil, '95', '5', '250', 'image_', '2.8', 'Badly split; no good spot', 'collected', '2019-06-09 08:26:01', nil],
          ['Blue53', '3', 'b', '3', 'Auto', 'Routine', 'SeMet', 'Luke', 'P2', '52.4 39.8 65.0 108.5', '1.5', '1.7', nil, '0.04', '0.25', '360', nil, '10', '300', nil, nil, 'Diff. center cutoff not met', 'skipped', '2019-06-09 08:29:01', nil],
          ['Blue53', '4', 'c', '3', 'Auto', 'Rods', 'Native', 'Luke', 'P21', '39 69.2 60 90 105.3', '1.5', '1.7', nil, '0.04', '0.25', '360', '95', '20', nil, 'image_', '1.6', nil, 'collected', nil, nil],
          ['Blue53', '5', 'd', '8', nil, 'Plates', nil, 'Luke', 'C222', '280 45 112 102 90', '1.5', '1.7', nil, '0.04', '0.25', '360', '95', '50', '300', 'image_', nil, nil, 'received', nil, nil]
        ],
        csv
      )
    end
  end
end
